@@ -19,7 +19,6 @@
 
 #include "ftp.h"
 
-static void try_bookmark_again (GtkWidget *widget, struct dialog_data *data);
 static void doadd_bookmark (GtkWidget *widget, struct dialog_data *data);
 static void bm_apply_changes (GtkWidget *widget, gpointer backup_data);
 static struct conn_categories *copy_bookmarks (struct conn_categories *bookmarks);
@@ -28,28 +27,35 @@ static void after_move (GtkCTree *ctree, GtkCTreeNode *child, GtkCTreeNode *pare
 	GtkCTreeNode *sibling, gpointer data);
 static gint bm_dblclick (GtkWidget *widget, GdkEventButton *event, gpointer data);
 static gint bm_enter (GtkWidget *widget, GdkEventKey *event, gpointer data);
-static gint hash_compare (gconstpointer path1, gconstpointer path2);
-static guint hash_function (gconstpointer key);
 static void new_folder_entry (gpointer data);
 static void do_make_new (GtkWidget *widget, struct dialog_data *data);
 static void new_item_entry (gpointer data);
 static void edit_entry (gpointer data);
 static void delete_entry (gpointer data);
+static void do_delete_entry (GtkWidget *widget, struct dialog_data *data);
 static void entry_apply_changes (GtkWidget *widget, struct conn_categories *entry);
 static void set_userpass_visible(GtkWidget *checkbutton, GtkWidget *entry);
 static void build_bookmarks_tree (void);
 static void clear_bookmarks_tree (void);
 
-static GtkWidget *bm_hostedit, *bm_portedit, *bm_diredit, *bm_useredit, 
-	*bm_passedit, *bm_acctedit, *anon_chk, *firewall_check, *bm_pathedit, 
-	*tree;
+static GtkWidget *bm_hostedit, *bm_portedit, *bm_localdiredit, *bm_remotediredit, 
+	*bm_useredit, *bm_passedit, *bm_acctedit, *anon_chk, *bm_pathedit, 
+	*bm_protocol, *tree;
 static struct conn_categories *new_bookmarks;
-static GHashTable *htable, /* Lookups by node */
-                  *new_bookmarks_htable; /* Lookups by path for new bookmarks */
+static GHashTable *new_bookmarks_htable; /* Lookups by path for new bookmarks */
 static GtkItemFactory *edit_factory;
 
 void run_bookmark (gpointer data) {
    struct conn_categories *tempentry;
+   int i;
+   
+   if (window2.hdata->tid > 0) {
+      pthread_kill (window2.hdata->tid, SIGINT);
+   }
+   while (window2.hdata->stopable) {
+      fix_display ();
+      sleep (0);
+   }
    
    tempentry = (struct conn_categories *) 
    	g_hash_table_lookup (bookmarks_htable, (char *) data);
@@ -63,13 +69,24 @@ void run_bookmark (gpointer data) {
       ftp_log (gftp_logging_misc, NULL, _("Bookmarks Error: There are some missing entries in this bookmark. Make sure you have a hostname and username\n"));
       return;
    }
-   else if (tempentry->pass == NULL || *tempentry->pass == '\0') {
-      MakeEditDialog (_("Enter Password"), _("Please enter your password for this site"), NULL, 0, 0,
-      	_("Connect"), try_bookmark_again, data, _("  Cancel  "), NULL, NULL);
-      return;
-   }
    
    if (window2.local != -1) disconnect (&window2);
+   if (tempentry->local_dir && *tempentry->local_dir != '\0') {
+      if (chdir (tempentry->local_dir) == 0) {
+         ftp_log (gftp_logging_misc, NULL, _("Successfully changed local directory to %s\n"), tempentry->local_dir);
+         gftp_set_directory (window1.hdata->ftpdata, tempentry->local_dir);
+         update_ftp_info (&window1);
+         gtk_clist_freeze (GTK_CLIST (window1.listbox));
+         delete_ftp_file_info (&window1);
+         add_local_files (&window1);
+         gtk_clist_thaw (GTK_CLIST (window1.listbox));
+      }
+      else {
+         ftp_log (gftp_logging_error, NULL, _("Could not change local directory to %s: %s\n"),
+         	tempentry->local_dir, g_strerror (errno));
+      }
+   }
+   
    gftp_set_username (window2.hdata->ftpdata, tempentry->user);
    if (strncmp (tempentry->pass, "@EMAIL@", 7) == 0) {
       gftp_set_password (window2.hdata->ftpdata, emailaddr);
@@ -81,59 +98,34 @@ void run_bookmark (gpointer data) {
       gftp_set_account (window2.hdata->ftpdata, tempentry->acct);
    }
    gftp_set_hostname (window2.hdata->ftpdata, tempentry->hostname);
-   gftp_set_directory (window2.hdata->ftpdata, tempentry->dir);
-   gftp_set_port (window2.hdata->ftpdata, strtol(tempentry->port, NULL, 10));
-   if (*tempentry->firewall == '1' && *proxy_config != '\0') {
-      if (firewall_host == NULL || *firewall_host == '\0') {
-         ftp_log (gftp_logging_error, NULL, _("Connect Error: You must specify a proxy hostname in the options dialog\n"));
-         return;
+   gftp_set_directory (window2.hdata->ftpdata, tempentry->remote_dir);
+   gftp_set_port (window2.hdata->ftpdata, tempentry->port);
+
+   for (i=0; gftp_protocols[i].name; i++) {
+      if (strcmp (gftp_protocols[i].name, tempentry->protocol) == 0) {
+         gftp_protocols[i].init (window2.hdata->ftpdata);
+         break;
       }
-      if (strcmp (proxy_config, "http") == 0) {
-         gftp_set_proxy_config (window2.hdata->ftpdata, "");
-         window2.hdata->protocol = http;
-      }
-      else {
-         gftp_set_proxy_config (window2.hdata->ftpdata, proxy_config);
-      }  
    }
-   else {
-      gftp_set_proxy_config (window2.hdata->ftpdata, "");
-      if (strcmp (proxy_config, "http") == 0) window2.hdata->protocol = ftp;
-   }
-   ftp_connect (window2.hdata, 1);
-}
-/*****************************************************************************/
-static void try_bookmark_again (GtkWidget *widget, struct dialog_data *data) {
-   struct conn_categories *tempentry;
-   gpointer d;
-   char *str;
+   if (!gftp_protocols[i].name) gftp_protocols[0].init (window2.hdata->ftpdata);
    
-   tempentry = (struct conn_categories *) 
-   	g_hash_table_lookup (bookmarks_htable, (char *) data->data);
-   str = gtk_entry_get_text (GTK_ENTRY (data->edit));
-   if (tempentry->pass) g_free (tempentry->pass);
-   tempentry->pass = g_malloc (strlen (str) + 1);
-   strcpy (tempentry->pass, str);
-   tempentry->save_password = 0;
-   d = data->data;
-   data->data = NULL;
-   gtk_widget_destroy (data->dialog);
-   run_bookmark (d);
+   ftp_connect (window2.hdata, 1);
 }
 /*****************************************************************************/
 void add_bookmark (gpointer data) {
    char *edttxt;
    
-   if (!check_status (_("Add Bookmark"), &window2, 0, 0, 0)) return;
+   if (!check_status (_("Add Bookmark"), &window2, 0, 0, 1)) return;
    
-   edttxt = gtk_entry_get_text (GTK_ENTRY (hostedit));
+   edttxt = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (hostedit)->entry));
    if (*edttxt == '\0') {
       ftp_log (gftp_logging_error, NULL, _("Add Bookmark: You must enter a hostname\n"));
       return;
    }
 
-   MakeEditDialog (_("Add Bookmark"), _("Enter the name of the bookmark you want to add\nYou can separate items by a / to put it into a submenu\n(ex: Linux Sites/Debian)"), NULL, 1, 1,
-   	_("Change"), doadd_bookmark, data,
+   MakeEditDialog (_("Add Bookmark"), _("Enter the name of the bookmark you want to add\nYou can separate items by a / to put it into a submenu\n(ex: Linux Sites/Debian)"), 
+   	NULL, 1, 1,
+   	_("Add"), doadd_bookmark, data,
    	_("  Cancel  "), NULL, NULL);
 }
 /*****************************************************************************/
@@ -168,19 +160,25 @@ static void doadd_bookmark (GtkWidget *widget, struct dialog_data *data) {
    tempentry->path = g_malloc (strlen (edttxt) + 1);
    strcpy (tempentry->path, edttxt);
 
-   edttxt = gtk_entry_get_text (GTK_ENTRY (hostedit));
+   edttxt = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (hostedit)->entry));
    tempentry->hostname = g_malloc (strlen (edttxt) + 1);
    strcpy (tempentry->hostname, edttxt);
 
-   edttxt = gtk_entry_get_text (GTK_ENTRY (portedit));
-   tempentry->port = g_malloc (strlen (edttxt) + 1);
-   strcpy (tempentry->port, edttxt);
+   edttxt = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (portedit)->entry));
+   tempentry->port = strtol (edttxt, NULL, 10);
+
+   tempentry->protocol = g_malloc (strlen (GFTP_GET_PROTOCOL_NAME (window2.hdata->ftpdata)) + 1);
+   strcpy (tempentry->protocol, GFTP_GET_PROTOCOL_NAME (window2.hdata->ftpdata));
+   
+   edttxt = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (window1.combo)->entry));
+   tempentry->local_dir = g_malloc (strlen (edttxt) + 1);
+   strcpy (tempentry->local_dir, edttxt);
 
    edttxt = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (window2.combo)->entry));
-   tempentry->dir = g_malloc (strlen (edttxt) + 1);
-   strcpy (tempentry->dir, edttxt);
+   tempentry->remote_dir = g_malloc (strlen (edttxt) + 1);
+   strcpy (tempentry->remote_dir, edttxt);
 
-   if ((edttxt = gtk_entry_get_text (GTK_ENTRY (useredit))) == NULL) {
+   if ((edttxt = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (useredit)->entry))) == NULL) {
       tempentry->user = g_malloc (strlen (ANON_LOGIN) + 1);
       strcpy (tempentry->user, ANON_LOGIN);
       
@@ -196,12 +194,9 @@ static void doadd_bookmark (GtkWidget *widget, struct dialog_data *data) {
       strcpy (tempentry->pass, edttxt);
    }
 
-   tempentry->firewall = g_malloc (1);
-   *tempentry->firewall = GTK_CHECK_MENU_ITEM (firewall_btn)->active;
-
    g_hash_table_insert (bookmarks_htable, tempentry->path, tempentry);
 
-   test.path = g_strconcat ("/Bookmarks/", tempentry->path, NULL);
+   test.path = g_strconcat (_("/Bookmarks/"), tempentry->path, NULL);
    gtk_item_factory_create_item (factory, &test, (gpointer) tempentry->path, 1);
    g_free (test.path);
    write_config_file ();
@@ -211,17 +206,16 @@ void edit_bookmarks (gpointer data) {
    GtkWidget *dialog, *scroll, *tempwid;
    GtkItemFactory *ifactory;
    GtkItemFactoryEntry dummy_item, menu_items[] = {
-      {"/_File",		NULL,   0,      	0, 	"<Branch>"},
-      {"/File/tearoff",  	NULL,   0,		0,	"<Tearoff>"},
-      {"/File/New Folder...", 	NULL, 	new_folder_entry,0},
-      {"/File/New Item...", 	NULL, 	new_item_entry,0},
-      {"/File/Delete", 		NULL, 	delete_entry,	0},
-      {"/File/Properties...", 	NULL, 	edit_entry,	0},
-      {"/File/sep",  		NULL,   0,		0,	"<Separator>"},
-      {"/File/Close",		NULL,	gtk_widget_destroy,	0}};
+      {N_("/_File"),		NULL,   0,      	0, 	"<Branch>"},
+      {N_("/File/tearoff"),  	NULL,   0,		0,	"<Tearoff>"},
+      {N_("/File/New Folder..."), NULL, new_folder_entry,0},
+      {N_("/File/New Item..."), NULL, 	new_item_entry,0},
+      {N_("/File/Delete"), 	NULL, 	delete_entry,	0},
+      {N_("/File/Properties..."), NULL, edit_entry,	0},
+      {N_("/File/sep"),  	NULL,   0,		0,	"<Separator>"},
+      {N_("/File/Close"),	NULL,	gtk_widget_destroy,	0}};
    int i;
    
-   htable = g_hash_table_new (hash_function, hash_compare);
    new_bookmarks = copy_bookmarks (hostcat);
    new_bookmarks_htable = build_hash_table (new_bookmarks);
    
@@ -233,14 +227,14 @@ void edit_bookmarks (gpointer data) {
    ifactory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<main>", NULL);
    create_item_factory (ifactory, 7, menu_items, NULL);
    create_item_factory (ifactory, 1, menu_items + 7, dialog);
-   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), ifactory->widget, FALSE, FALSE, FALSE);
+   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), ifactory->widget, FALSE, FALSE, 0);
    gtk_widget_show (ifactory->widget);
    
    scroll = gtk_scrolled_window_new (NULL, NULL);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroll), 
    	GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    gtk_widget_set_usize (scroll, 150, 200);
-   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), scroll, TRUE, TRUE, TRUE);
+   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), scroll, TRUE, TRUE, 0);
    gtk_container_border_width (GTK_CONTAINER (scroll), 3);
    gtk_widget_show (scroll);
 
@@ -263,21 +257,21 @@ void edit_bookmarks (gpointer data) {
    build_bookmarks_tree ();
 
    tempwid = gtk_button_new_with_label (_("OK"));
-   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, FALSE);
-   gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (bm_apply_changes), (gpointer) NULL);
+   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, 0);
+   gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (bm_apply_changes), NULL);
    gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (bm_close_dialog), (gpointer) dialog);
    GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
    gtk_widget_show (tempwid);
 
    tempwid = gtk_button_new_with_label (_("  Cancel  "));
-   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, FALSE);
+   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, 0);
    gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (bm_close_dialog), (gpointer) dialog);
    GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
    gtk_widget_grab_focus (tempwid);
    gtk_widget_show (tempwid);
 
    tempwid = gtk_button_new_with_label (_("Apply"));
-   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, FALSE);
+   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, 0);
    gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (bm_apply_changes), (gpointer) 1);
    GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
    gtk_widget_show (tempwid);
@@ -287,14 +281,13 @@ void edit_bookmarks (gpointer data) {
 /*****************************************************************************/
 static void bm_apply_changes (GtkWidget *widget, gpointer backup_data) {
    struct conn_categories *tempentry, *delentry;
-   char tempstr[MAXSTR];
+   char *tempstr;
    
    tempentry = hostcat->children;
    while (tempentry != NULL) {
       if (tempentry->path) {
          if (!tempentry->isfolder) {
-            g_snprintf (tempstr, sizeof (tempstr), _("/Bookmarks/%s"), tempentry->path);
-            tempstr[sizeof (tempstr) - 1] = '\0';
+            tempstr = g_strdup_printf (_("/Bookmarks/%s"), tempentry->path);
 
 /* The gtk_item_factory_delete_item () has a bug in it. This will be fixed in
    version 1.2.4 of gtk+. I will uncomment these in my next version */
@@ -302,15 +295,14 @@ static void bm_apply_changes (GtkWidget *widget, gpointer backup_data) {
 /*FIXME            gtk_item_factory_delete_item (factory, tempstr);*/
             gtk_widget_destroy (gtk_item_factory_get_item (factory, tempstr));
             g_free (tempentry->path);
+            g_free (tempstr);
          }
       }
       if (tempentry->hostname) g_free (tempentry->hostname);
-      if (tempentry->dir) g_free (tempentry->dir);
+      if (tempentry->remote_dir) g_free (tempentry->remote_dir);
       if (tempentry->user) g_free (tempentry->user);
       if (tempentry->pass) g_free (tempentry->pass);
       if (tempentry->acct) g_free (tempentry->acct);
-      if (tempentry->port) g_free (tempentry->port);
-      if (tempentry->firewall) g_free (tempentry->firewall);
 
       if (tempentry->children != NULL) {
          tempentry = tempentry->children;
@@ -321,22 +313,22 @@ static void bm_apply_changes (GtkWidget *widget, gpointer backup_data) {
                delentry = tempentry;
                tempentry = tempentry->prev;
                if (delentry->isfolder) {
-                  g_snprintf (tempstr, sizeof (tempstr), _("/Bookmarks/%s"), delentry->path);
-                  tempstr[sizeof (tempstr) - 1] = '\0';
+                  tempstr = g_strdup_printf (_("/Bookmarks/%s"), delentry->path);
 /*FIXME                  gtk_item_factory_delete_item (factory, tempstr);*/
                   gtk_widget_destroy (gtk_item_factory_get_item (factory, tempstr));
                   g_free (delentry->path);
+                  g_free (tempstr);
                }
                g_free (delentry);
             }
             delentry = tempentry;
             tempentry = tempentry->next;
             if (delentry->isfolder && tempentry != NULL) {
-               g_snprintf (tempstr, sizeof (tempstr), _("/Bookmarks/%s"), delentry->path);
-               tempstr[sizeof (tempstr) - 1] = '\0';
+               tempstr = g_strdup_printf (_("/Bookmarks/%s"), delentry->path);
 /*FIXME               gtk_item_factory_delete_item (factory, tempstr);*/
                gtk_widget_destroy (gtk_item_factory_get_item (factory, tempstr));
                g_free (delentry->path);
+               g_free (tempstr);
                g_free (delentry);
             }
          }
@@ -360,6 +352,7 @@ static void bm_apply_changes (GtkWidget *widget, gpointer backup_data) {
       new_bookmarks_htable = NULL;
    }
    build_bookmarks_menu ();
+   write_config_file ();
 }
 /*****************************************************************************/
 static struct conn_categories *copy_bookmarks (struct conn_categories *bookmarks) {
@@ -376,6 +369,7 @@ static struct conn_categories *copy_bookmarks (struct conn_categories *bookmarks
    while (tempentry != NULL) {
       newentry = g_malloc0 (sizeof (struct conn_categories));
       newentry->isfolder = tempentry->isfolder;
+      newentry->save_password = tempentry->save_password;
       if (tempentry->path) {
          newentry->path = g_malloc (strlen (tempentry->path) + 1);
          strcpy (newentry->path, tempentry->path);
@@ -384,9 +378,17 @@ static struct conn_categories *copy_bookmarks (struct conn_categories *bookmarks
          newentry->hostname = g_malloc (strlen (tempentry->hostname) + 1);
          strcpy (newentry->hostname, tempentry->hostname);
       }
-      if (tempentry->dir) {
-         newentry->dir = g_malloc (strlen (tempentry->dir) + 1);
-         strcpy (newentry->dir, tempentry->dir);
+      if (tempentry->protocol) {
+         newentry->protocol = g_malloc (strlen (tempentry->protocol) + 1);
+         strcpy (newentry->protocol, tempentry->protocol);
+      }
+      if (tempentry->local_dir) {
+         newentry->local_dir = g_malloc (strlen (tempentry->local_dir) + 1);
+         strcpy (newentry->local_dir, tempentry->local_dir);
+      }
+      if (tempentry->remote_dir) {
+         newentry->remote_dir = g_malloc (strlen (tempentry->remote_dir) + 1);
+         strcpy (newentry->remote_dir, tempentry->remote_dir);
       }
       if (tempentry->user) {
          newentry->user = g_malloc (strlen (tempentry->user) + 1);
@@ -400,14 +402,7 @@ static struct conn_categories *copy_bookmarks (struct conn_categories *bookmarks
          newentry->acct = g_malloc (strlen (tempentry->acct) + 1);
          strcpy (newentry->acct, tempentry->acct);
       }
-      if (tempentry->port) {
-         newentry->port = g_malloc (strlen (tempentry->port) + 1);
-         strcpy (newentry->port, tempentry->port);
-      }
-      if (tempentry->firewall) {
-         newentry->firewall = g_malloc (1);
-         *newentry->firewall = *tempentry->firewall;
-      }
+      newentry->port = tempentry->port;
       if (sibling == NULL) {
          if (preventry->children == NULL) preventry->children = newentry;
          else {
@@ -465,11 +460,11 @@ static void after_move (GtkCTree *ctree, GtkCTreeNode *child, GtkCTreeNode *pare
 	GtkCTreeNode *sibling, gpointer data) {
   
    struct conn_categories *childentry, *siblingentry, *parententry, *tempentry;
-   char tempstr[MAXSTR], *pos, *stpos;
+   char *tempstr, *pos, *stpos;
    
-   childentry = (struct conn_categories *) g_hash_table_lookup (htable, child);
-   parententry = (struct conn_categories *) g_hash_table_lookup (htable, parent);
-   siblingentry = (struct conn_categories *) g_hash_table_lookup (htable, sibling);
+   childentry = gtk_ctree_node_get_row_data (ctree, child);
+   parententry = gtk_ctree_node_get_row_data (ctree, parent);
+   siblingentry = gtk_ctree_node_get_row_data (ctree, sibling);
 
    tempentry = childentry->prev->children;
    if (tempentry == childentry) {
@@ -502,12 +497,12 @@ static void after_move (GtkCTree *ctree, GtkCTreeNode *child, GtkCTreeNode *pare
          g_hash_table_remove (new_bookmarks_htable, tempentry->path);
          if ((pos = strrchr (tempentry->path, '/')) == NULL) pos = tempentry->path;
          else pos++;
-         g_snprintf (tempstr, sizeof (tempstr), "%s/%s", tempentry->prev->path, pos);
-         tempstr[sizeof (tempstr) - 1] = '\0';
+         tempstr = g_strdup_printf ("%s/%s", tempentry->prev->path, pos);
          for (stpos = tempstr; *stpos == '/'; stpos++);
          g_free (tempentry->path);
          tempentry->path = g_malloc (strlen (stpos) + 1);
          strcpy (tempentry->path, stpos);
+         g_free (tempstr);
          g_hash_table_insert (new_bookmarks_htable, tempentry->path, tempentry);
          if (tempentry->children != NULL) tempentry = tempentry->children;
          else {
@@ -550,23 +545,15 @@ static gint bm_enter (GtkWidget *widget, GdkEventKey *event, gpointer data) {
    return (TRUE);
 }
 /*****************************************************************************/
-static gint hash_compare (gconstpointer path1, gconstpointer path2) {
-   return (path1 == path2);
-}
-/*****************************************************************************/
-static guint hash_function (gconstpointer key) {
-   return ((guint) key * 33);
-}
-/*****************************************************************************/
 static void new_folder_entry (gpointer data) {
    MakeEditDialog (_("New Folder"), _("Enter the name of the new folder to create"), NULL, 1, 1,
-   	_("Create"), do_make_new, (void *) 1,
+   	_("Create"), do_make_new, (gpointer) 1,
    	_("  Cancel  "), NULL, NULL);
 }
 /*****************************************************************************/
 static void new_item_entry (gpointer data) {
    MakeEditDialog (_("New Folder"), _("Enter the name of the new item to create"), NULL, 1, 1,
-   	_("Create"), do_make_new, 0,
+   	_("Create"), do_make_new, (gpointer) 0,
    	_("  Cancel  "), NULL, NULL);
 }
 /*****************************************************************************/
@@ -583,7 +570,7 @@ static void do_make_new (GtkWidget *widget, struct dialog_data *data) {
    newentry->path = g_malloc (strlen (str) + 1);
    strcpy (newentry->path, str);
    newentry->prev = new_bookmarks;
-   if (data) newentry->isfolder = 1;
+   if (data->data) newentry->isfolder = 1;
    
    if (new_bookmarks->children == NULL) {
       new_bookmarks->children = newentry;
@@ -607,17 +594,18 @@ static void do_make_new (GtkWidget *widget, struct dialog_data *data) {
    		new_bookmarks->cnode, NULL, text, 5, NULL, NULL, 
    		NULL, NULL, FALSE, FALSE);
    }
-   g_hash_table_insert (htable, newentry->cnode, newentry);
+   gtk_ctree_node_set_row_data (GTK_CTREE (tree), newentry->cnode, newentry);
    g_hash_table_insert (new_bookmarks_htable, newentry->path, newentry);
 }
 /*****************************************************************************/
 static void edit_entry (gpointer data) {
-   GtkWidget *table, *tempwid, *dialog;
+   GtkWidget *table, *tempwid, *dialog, *menu;
    struct conn_categories *entry;
+   int i, num;
    char *pos;
    
    if (GTK_CLIST (tree)->selection == NULL) return;
-   entry = (struct conn_categories *) g_hash_table_lookup (htable, GTK_CLIST (tree)->selection->data);
+   entry = gtk_ctree_node_get_row_data (GTK_CTREE (tree), GTK_CLIST (tree)->selection->data);
    if (entry == NULL || entry == new_bookmarks) return;
 
    dialog = gtk_dialog_new ();
@@ -627,10 +615,10 @@ static void edit_entry (gpointer data) {
    gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
    
    tempwid = gtk_frame_new (NULL);
-   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), tempwid, TRUE, TRUE, FALSE);
+   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), tempwid, TRUE, TRUE, 0);
    gtk_widget_show (tempwid);
    
-   table = gtk_table_new (10, 2, FALSE);
+   table = gtk_table_new (11, 2, FALSE);
    gtk_container_border_width (GTK_CONTAINER (table), 5);
    gtk_table_set_row_spacings (GTK_TABLE (table), 5);
    gtk_table_set_col_spacings (GTK_TABLE (table), 5);
@@ -668,42 +656,82 @@ static void edit_entry (gpointer data) {
    bm_portedit = gtk_entry_new ();
    gtk_table_attach_defaults (GTK_TABLE (table), bm_portedit, 1, 2, 2, 3);
    if (entry->isfolder) gtk_widget_set_sensitive (bm_portedit, 0);
-   else if (entry->port) gtk_entry_set_text (GTK_ENTRY (bm_portedit), entry->port);
+   else if (entry->port) {
+      pos = g_strdup_printf ("%d", entry->port);
+      gtk_entry_set_text (GTK_ENTRY (bm_portedit), pos);
+      g_free (pos);
+   }
    gtk_widget_show (bm_portedit);
 
-   tempwid = gtk_label_new (_("Directory:"));
+   tempwid = gtk_label_new (_("Protocol:"));
    gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
    gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 3, 4);
    gtk_widget_show (tempwid);
 
-   bm_diredit = gtk_entry_new ();
-   gtk_table_attach_defaults (GTK_TABLE (table), bm_diredit, 1, 2, 3, 4);
-   if (entry->isfolder) gtk_widget_set_sensitive (bm_diredit, 0);
-   else if (entry->dir) gtk_entry_set_text (GTK_ENTRY (bm_diredit), entry->dir);
-   gtk_widget_show (bm_diredit);
+   menu = gtk_option_menu_new ();
+   gtk_table_attach_defaults (GTK_TABLE (table), menu, 1, 2, 3, 4);
+   gtk_widget_show (menu);
 
-   tempwid = gtk_hseparator_new ();
-   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 2, 4, 5);
+   gtk_option_menu_set_history (GTK_OPTION_MENU (menu), 0);
+
+   num = 0;
+   bm_protocol = gtk_menu_new ();
+   for (i=0; gftp_protocols[i].name; i++) {
+      tempwid = gtk_menu_item_new_with_label (gftp_protocols[i].name);
+      gtk_object_set_user_data (GTK_OBJECT (tempwid), gftp_protocols[i].name);
+      gtk_menu_append (GTK_MENU (bm_protocol), tempwid);
+      gtk_widget_show (tempwid);
+      if (entry->protocol && strcmp (gftp_protocols[i].name, entry->protocol) == 0) {
+         num = i;
+      }
+   }
+   gtk_option_menu_set_menu (GTK_OPTION_MENU (menu), bm_protocol);
+   gtk_option_menu_set_history (GTK_OPTION_MENU (menu), num);
+   
+   tempwid = gtk_label_new (_("Remote Directory:"));
+   gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
+   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 4, 5);
    gtk_widget_show (tempwid);
 
-   tempwid = gtk_label_new (_("Username:"));
+   bm_remotediredit = gtk_entry_new ();
+   gtk_table_attach_defaults (GTK_TABLE (table), bm_remotediredit, 1, 2, 4, 5);
+   if (entry->isfolder) gtk_widget_set_sensitive (bm_remotediredit, 0);
+   else if (entry->remote_dir) gtk_entry_set_text (GTK_ENTRY (bm_remotediredit), entry->remote_dir);
+   gtk_widget_show (bm_remotediredit);
+   
+   tempwid = gtk_label_new (_("Local Directory:"));
    gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
    gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 5, 6);
    gtk_widget_show (tempwid);
 
+   bm_localdiredit = gtk_entry_new ();
+   gtk_table_attach_defaults (GTK_TABLE (table), bm_localdiredit, 1, 2, 5, 6);
+   if (entry->isfolder) gtk_widget_set_sensitive (bm_localdiredit, 0);
+   else if (entry->local_dir) gtk_entry_set_text (GTK_ENTRY (bm_localdiredit), entry->local_dir);
+   gtk_widget_show (bm_localdiredit);
+
+   tempwid = gtk_hseparator_new ();
+   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 2, 6, 7);
+   gtk_widget_show (tempwid);
+
+   tempwid = gtk_label_new (_("Username:"));
+   gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
+   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 7, 8);
+   gtk_widget_show (tempwid);
+
    bm_useredit = gtk_entry_new ();
-   gtk_table_attach_defaults (GTK_TABLE (table), bm_useredit, 1, 2, 5, 6);
+   gtk_table_attach_defaults (GTK_TABLE (table), bm_useredit, 1, 2, 7, 8);
    if (entry->isfolder) gtk_widget_set_sensitive (bm_useredit, 0);
    else if (entry->user) gtk_entry_set_text (GTK_ENTRY (bm_useredit), entry->user);
    gtk_widget_show (bm_useredit);
 
    tempwid = gtk_label_new (_("Password:"));
    gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
-   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 6, 7);
+   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 8, 9);
    gtk_widget_show (tempwid);
 
    bm_passedit = gtk_entry_new ();
-   gtk_table_attach_defaults (GTK_TABLE (table), bm_passedit, 1, 2, 6, 7);
+   gtk_table_attach_defaults (GTK_TABLE (table), bm_passedit, 1, 2, 8, 9);
    gtk_entry_set_visibility (GTK_ENTRY (bm_passedit), FALSE);
    if (entry->isfolder) gtk_widget_set_sensitive (bm_passedit, 0);
    else if (entry->pass) gtk_entry_set_text (GTK_ENTRY (bm_passedit), entry->pass);
@@ -711,17 +739,17 @@ static void edit_entry (gpointer data) {
 
    tempwid = gtk_label_new (_("Account:"));
    gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
-   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 7, 8);
+   gtk_table_attach_defaults (GTK_TABLE (table), tempwid, 0, 1, 9, 10);
    gtk_widget_show (tempwid);
 
    bm_acctedit = gtk_entry_new ();
-   gtk_table_attach_defaults (GTK_TABLE (table), bm_acctedit, 1, 2, 7, 8);
+   gtk_table_attach_defaults (GTK_TABLE (table), bm_acctedit, 1, 2, 9, 10);
    if (entry->isfolder) gtk_widget_set_sensitive (bm_acctedit, 0);
    else if (entry->acct) gtk_entry_set_text (GTK_ENTRY (bm_acctedit), entry->acct);
    gtk_widget_show (bm_acctedit);
 
    anon_chk = gtk_check_button_new_with_label (_("Log in as ANONYMOUS"));
-   gtk_table_attach_defaults (GTK_TABLE (table), anon_chk, 0, 2, 8, 9);
+   gtk_table_attach_defaults (GTK_TABLE (table), anon_chk, 0, 2, 10, 11);
    if (entry->isfolder) gtk_widget_set_sensitive (anon_chk, 0);
    else {
       gtk_signal_connect (GTK_OBJECT (anon_chk), "toggled", GTK_SIGNAL_FUNC (set_userpass_visible), NULL);
@@ -729,33 +757,22 @@ static void edit_entry (gpointer data) {
    }
    gtk_widget_show (anon_chk);
 
-   firewall_check = gtk_check_button_new_with_label (_("Connect through FTP Proxy Server"));
-   gtk_table_attach_defaults (GTK_TABLE (table), firewall_check, 0, 2, 9, 10);
-   if (*proxy_config != '\0' && !entry->isfolder) {
-      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (firewall_check), entry->firewall && entry->firewall[0] == '1');
-   }
-   else {
-      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (firewall_check), 1);
-      gtk_widget_set_sensitive (firewall_check, 0);
-   }
-   gtk_widget_show (firewall_check);
-   
    tempwid = gtk_button_new_with_label (_("OK"));
-   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, FALSE);
+   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, 0);
    gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (entry_apply_changes), (gpointer) entry);
    gtk_signal_connect_object (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (dialog));
    GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
    gtk_widget_show (tempwid);
 
    tempwid = gtk_button_new_with_label (_("  Cancel  "));
-   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, FALSE);
+   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, 0);
    gtk_signal_connect_object (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (dialog));
    GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
    gtk_widget_grab_focus (tempwid);
    gtk_widget_show (tempwid);
 
    tempwid = gtk_button_new_with_label (_("Apply"));
-   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, FALSE);
+   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, 0);
    gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (entry_apply_changes), (gpointer) entry);
    GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
    gtk_widget_show (tempwid);
@@ -764,13 +781,31 @@ static void edit_entry (gpointer data) {
 }
 /*****************************************************************************/
 static void delete_entry (gpointer data) {
-   struct conn_categories *entry, *tempentry, *delentry;
+   struct conn_categories *entry;
+   struct dialog_data d_data;
+   char *tempstr, *pos;
    
    if (GTK_CLIST (tree)->selection == NULL) return;
-   entry = (struct conn_categories *) g_hash_table_lookup (htable, GTK_CLIST (tree)->selection->data);
+   entry = gtk_ctree_node_get_row_data (GTK_CTREE (tree), GTK_CLIST (tree)->selection->data);
    if (entry == NULL || entry->prev == NULL) return;
 
-   g_hash_table_remove (htable, entry->cnode);
+   d_data.data = entry;
+   if (!entry->isfolder) do_delete_entry (NULL, &d_data);
+   else {
+      if ((pos = strrchr (entry->path, '/')) == NULL) pos = entry->path;
+      else pos++;
+      tempstr = g_strdup_printf (_("Are you sure you want to erase the bookmark\n%s and all it's children?"), pos);
+      MakeYesNoDialog (_("Delete Bookmark"), tempstr, 1, 2,
+      	_("Yes"), do_delete_entry, entry,
+      	_("No"), NULL, NULL);
+   }
+}
+/*****************************************************************************/
+static void do_delete_entry (GtkWidget *widget, struct dialog_data *data) {
+   struct conn_categories *entry, *tempentry, *delentry;
+   
+   entry = (struct conn_categories *) data->data;
+
    g_hash_table_remove (new_bookmarks_htable, entry->path);
    gtk_ctree_remove_node (GTK_CTREE (tree), entry->cnode);
    if (entry->prev->children == entry) {
@@ -788,11 +823,9 @@ static void delete_entry (gpointer data) {
    while (tempentry != NULL) {
       if (tempentry->path) g_free (tempentry->path);
       if (tempentry->hostname) g_free (tempentry->hostname);
-      if (tempentry->dir) g_free (tempentry->dir);
+      if (tempentry->remote_dir) g_free (tempentry->remote_dir);
       if (tempentry->user) g_free (tempentry->user);
       if (tempentry->pass) g_free (tempentry->pass);
-      if (tempentry->port) g_free (tempentry->port);
-      if (tempentry->firewall) g_free (tempentry->firewall);
 
       if (tempentry->children != NULL) {
          tempentry = tempentry->children;
@@ -812,6 +845,7 @@ static void delete_entry (gpointer data) {
 static void entry_apply_changes (GtkWidget *widget, struct conn_categories *entry) {
    struct conn_categories *tempentry, *nextentry;
    char *str, *pos, *newpath, tempchar;
+   GtkWidget *tempwid;
    int oldpathlen;
 
    oldpathlen = strlen (entry->path);
@@ -835,14 +869,23 @@ static void entry_apply_changes (GtkWidget *widget, struct conn_categories *entr
    strcpy (entry->hostname, str);
    
    str = gtk_entry_get_text (GTK_ENTRY (bm_portedit));
-   if (entry->port) g_free (entry->port);
-   entry->port = g_malloc (strlen (str) + 1);
-   strcpy (entry->port, str);
+   entry->port = strtol (str, NULL, 10);
    
-   str = gtk_entry_get_text (GTK_ENTRY (bm_diredit));
-   if (entry->dir) g_free (entry->dir);
-   entry->dir = g_malloc (strlen (str) + 1);
-   strcpy (entry->dir, str);
+   tempwid = gtk_menu_get_active (GTK_MENU (bm_protocol));
+   str = gtk_object_get_user_data (GTK_OBJECT (tempwid));
+   if (entry->protocol) g_free (entry->protocol);
+   entry->protocol = g_malloc (strlen (str) + 1);
+   strcpy (entry->protocol, str);
+   
+   str = gtk_entry_get_text (GTK_ENTRY (bm_remotediredit));
+   if (entry->remote_dir) g_free (entry->remote_dir);
+   entry->remote_dir = g_malloc (strlen (str) + 1);
+   strcpy (entry->remote_dir, str);
+
+   str = gtk_entry_get_text (GTK_ENTRY (bm_localdiredit));
+   if (entry->local_dir) g_free (entry->local_dir);
+   entry->local_dir = g_malloc (strlen (str) + 1);
+   strcpy (entry->local_dir, str);
    
    if (GTK_TOGGLE_BUTTON (anon_chk)->active) str = ANON_LOGIN;
    else str = gtk_entry_get_text (GTK_ENTRY (bm_useredit));
@@ -857,14 +900,12 @@ static void entry_apply_changes (GtkWidget *widget, struct conn_categories *entr
    strcpy (entry->pass, str);
    entry->save_password = *entry->pass != '\0';
    
-   str = gtk_entry_get_text (GTK_ENTRY (bm_acctedit));
+   if (GTK_TOGGLE_BUTTON (anon_chk)->active) str = "";
+   else str = gtk_entry_get_text (GTK_ENTRY (bm_acctedit));
    if (entry->acct) g_free (entry->acct);
    entry->acct = g_malloc (strlen (str) + 1);
    strcpy (entry->acct, str);
    
-   if (!entry->firewall) entry->firewall = g_malloc (1);
-   *entry->firewall = GTK_TOGGLE_BUTTON (firewall_check)->active ? '1' : '0';
-
    if (strcmp (entry->path, newpath) != 0) {
       tempentry = entry;
       nextentry = entry->next;
@@ -893,6 +934,7 @@ static void entry_apply_changes (GtkWidget *widget, struct conn_categories *entr
 static void set_userpass_visible(GtkWidget *checkbutton, GtkWidget *entry) {
    gtk_widget_set_sensitive (bm_useredit, !GTK_TOGGLE_BUTTON (anon_chk)->active);
    gtk_widget_set_sensitive (bm_passedit, !GTK_TOGGLE_BUTTON (anon_chk)->active);
+   gtk_widget_set_sensitive (bm_acctedit, !GTK_TOGGLE_BUTTON (anon_chk)->active);   
 }
 /*****************************************************************************/
 void build_bookmarks_menu (void) {
@@ -929,8 +971,8 @@ static void build_bookmarks_tree (void) {
    text[0] = text[1] = _("Bookmarks");
    parent = gtk_ctree_insert_node (GTK_CTREE (tree), NULL, NULL, text, 5, 
    	dir_pixmap, dir_mask, open_dir_pixmap, open_dir_mask, FALSE, TRUE);
+   gtk_ctree_node_set_row_data (GTK_CTREE (tree), parent, new_bookmarks);
    new_bookmarks->cnode = parent;
-   g_hash_table_insert (htable, new_bookmarks->cnode, new_bookmarks);
    
    tempentry = new_bookmarks->children;
    while (tempentry != NULL) {
@@ -954,7 +996,7 @@ static void build_bookmarks_tree (void) {
                preventry->cnode = gtk_ctree_insert_node (GTK_CTREE (tree), 
                		preventry->prev->cnode, NULL, text, 5, dir_pixmap, dir_mask, 
                		open_dir_pixmap, open_dir_mask, FALSE, FALSE);
-               g_hash_table_insert (htable, preventry->cnode, preventry);
+               gtk_ctree_node_set_row_data (GTK_CTREE (tree), preventry->cnode, preventry);
             }
             g_free (str);
             pos++;
@@ -968,7 +1010,7 @@ static void build_bookmarks_tree (void) {
       tempentry->cnode = gtk_ctree_insert_node (GTK_CTREE (tree), 
       		tempentry->prev->cnode, NULL, text, 5, NULL, NULL, 
       		NULL, NULL, FALSE, FALSE);
-      g_hash_table_insert (htable, tempentry->cnode, tempentry);
+      gtk_ctree_node_set_row_data (GTK_CTREE (tree), tempentry->cnode, tempentry);
 
       while (tempentry->next == NULL && tempentry->prev != NULL) tempentry = tempentry->prev;
       tempentry = tempentry->next;
@@ -993,7 +1035,5 @@ static void clear_bookmarks_tree (void) {
       tempentry->cnode = NULL;
       tempentry = tempentry->next;
    }
-   g_hash_table_destroy (htable);
-   htable = g_hash_table_new (hash_function, hash_compare);
 }
 /*****************************************************************************/
