@@ -19,23 +19,26 @@
 
 #include "ftp.h"
 
-char *insert_commas (unsigned long number, char *dest, int size) {
-   char *frompos, *topos, *src;
-   int num, rem, i;
+char *insert_commas (unsigned long number) {
+   char *frompos, *topos, *src, *dest;
+   unsigned long tempnum;
+   int len, num, rem, i;
    
-   src = g_malloc (size);
-   g_snprintf (src, size, "%ld", number);
-   src[size-1] = '\0';
+   len = 0;
+   tempnum = number;
+   while (tempnum > 0) {
+      tempnum /= 10;
+      len++;
+   }
+   if (len % 3 > 0) len++;
+   len += len / 3;
+   if (len <= 0) return (NULL);
+   
+   src = g_strdup_printf ("%ld", number);
+   dest = g_malloc (len);
 
    num = strlen (src) / 3 - 1;
    rem = strlen (src) % 3;
-   if (strlen (src) + num + (rem == 0 ? 0 : 1) + 1 > size) {
-      if (strlen (src) < size) strncpy (dest, src, size);
-      else memset (dest, 'X', size-1);
-      dest[size-1] = '\0';
-      g_free (src);
-      return (dest);
-   }
    frompos = src;
    topos = dest;
    for (i=0; i<rem; i++) *topos++ = *frompos++;
@@ -57,22 +60,22 @@ void add_local_files (struct ftp_window_data *wdata) {
    char tempstr[MAXSTR];
 
    wdata->local = 1;
-   wdata->flags &= ~(WINDOW_CACHED);
-   wdata->totalitems = wdata->numselected = 0;
-   wdata->flags &= ~(WINDOW_SORTED);
+   wdata->cached = 0;
+   wdata->hdata->totalfiles = wdata->numselected = 0;
+   wdata->sorted = 0;
    gftp_set_directory (wdata->hdata->ftpdata, getcwd (tempstr, sizeof (tempstr)));
-   if ((wdata->hdata->files = get_local_files (NULL, &wdata->totalitems)) == NULL) {
+   if ((wdata->hdata->files = get_local_files (NULL, &wdata->hdata->totalfiles)) == NULL) {
       wdata->local = -1;
-      wdata->totalitems = wdata->numselected = 0;
+      wdata->hdata->totalfiles = wdata->numselected = 0;
    }
    else sortrows (GTK_CLIST(wdata->listbox), wdata->sortcol, (gpointer) wdata);
    gtk_clist_select_row (GTK_CLIST (wdata->listbox), 0, 0);
    update_ftp_info (wdata);
 }
 /*****************************************************************************/
-struct ftp_file_data *get_local_files (char *path, int *total) {
-   struct ftp_file_data *files;
+GList *get_local_files (char *path, int *total) {
    char curdir[MAXSTR];
+   GList *files;
    FILE *dir;
    
    *total = 0;
@@ -81,106 +84,94 @@ struct ftp_file_data *get_local_files (char *path, int *total) {
       if (chdir(path) == -1) return (NULL);
    }
    dir = popen ("/bin/ls -al", "r");
-   files = parse_local_file (dir, 1, total, ftp);
+   files = parse_local_file (dir, 1, total);
    pclose (dir);
    if (path != NULL) chdir (curdir);
    return (files);
 }
 /*****************************************************************************/
-struct ftp_file_data *parse_local_file (FILE *fd, int local, int *total, protocol_type proto) {
-   struct ftp_file_data *files, *newfle, *lastfle;
-   char tempstr[MAXSTR];
-   int isdotdot = 0;
+GList *parse_local_file (FILE *fd, int local, int *total) {
+   struct ftp_file_data *newfle;
    struct stat st;
    gftp_file fle;
+   GList *files;
 
-   lastfle = files = NULL;
-   while (fgets (tempstr, sizeof (tempstr)-1, fd)) {
-      if (tempstr[strlen(tempstr)-1] == '\n') tempstr[strlen(tempstr)-1] = '\0';
-      if (tempstr[strlen(tempstr)-1] == '\r') tempstr[strlen(tempstr)-1] = '\0';
+   files = NULL;
+   while (gftp_get_next_file (window1.hdata->ftpdata, &fle, fd)) {
+      if (fle.file == NULL) break;
+      if (strcmp (fle.file, ".") == 0 || strcmp (fle.file, "..") == 0) continue;
 
-      if (proto == ftp ? gftp_parse_ls (tempstr, &fle) == 0 : parse_html_line (tempstr, &fle)) {
-         if (strcmp (fle.file, ".") == 0) continue;
-         if (strcmp(fle.file, "..") == 0) isdotdot = 1;
-
-         newfle = g_malloc0 (sizeof (struct ftp_file_data));
-         memcpy (newfle, &fle, sizeof (gftp_file));
-         newfle->flags = 0;
-         if (local) {
-            stat (newfle->file, &st);
-            if (S_ISDIR (st.st_mode)) newfle->flags |= FILE_ISDIR;
-            if (newfle->attribs[0] == 'l') newfle->flags |= FILE_ISLINK;
-            if ((strchr (newfle->attribs, 'x') != NULL) && !(newfle->flags & FILE_ISDIR)
-            	&& !(newfle->flags & FILE_ISLINK)) newfle->flags |= FILE_ISEXE;
-         }
-         else {
-            if (strchr (newfle->attribs, 'd') != NULL) newfle->flags |= FILE_ISDIR;
-            if (strchr (newfle->attribs, 'l') != NULL) newfle->flags |= FILE_ISLINK;
-            if ((strchr (newfle->attribs, 'x') != NULL) && !(newfle->flags & FILE_ISDIR)
-            	&& !(newfle->flags & FILE_ISLINK)) newfle->flags |= FILE_ISEXE;
-         }
-         newfle->next = NULL;
-         if (lastfle == NULL) files = newfle;
-         else lastfle->next = newfle;
-         lastfle = newfle;  
-         (*total)++;
-      }
-   }
-   if (!isdotdot) {
       newfle = g_malloc0 (sizeof (struct ftp_file_data));
-      newfle->file = g_malloc (3);
-      strcpy (newfle->file, "..");
-      newfle->user = g_malloc (1);
-      *newfle->user = '\0';
-      newfle->group = g_malloc (1);
-      *newfle->group = '\0';
-      newfle->attribs = g_malloc (1);
-      *newfle->attribs = '\0';
-      
-      newfle->next = NULL;
-      newfle->flags = FILE_ISDIR;
-      if (lastfle == NULL) files = newfle;
-      else lastfle->next = newfle;
-      lastfle = newfle;  
+      memcpy (newfle, &fle, sizeof (gftp_file));
+      if (local) {
+         stat (newfle->file, &st);
+         if (S_ISDIR (st.st_mode)) newfle->isdir = 1;
+         if (newfle->attribs[0] == 'l') newfle->islink = 1;
+         if ((strchr (newfle->attribs, 'x') != NULL) && !newfle->isdir
+         	&& !newfle->islink) newfle->isexe = 1;
+      }
+      else {
+         if (strchr (newfle->attribs, 'd') != NULL) newfle->isdir = 1;
+         if (strchr (newfle->attribs, 'l') != NULL) newfle->islink = 1;
+         if ((strchr (newfle->attribs, 'x') != NULL) && !newfle->isdir
+         	&& !newfle->islink) newfle->isexe = 1;
+      }
+      files = g_list_append (files, newfle);
       (*total)++;
    }
+   
+   newfle = g_malloc0 (sizeof (struct ftp_file_data));
+   newfle->file = g_malloc (3);
+   strcpy (newfle->file, "..");
+   newfle->user = g_malloc (1);
+   *newfle->user = '\0';
+   newfle->group = g_malloc (1);
+   *newfle->group = '\0';
+   newfle->attribs = g_malloc (1);
+   *newfle->attribs = '\0';
+   newfle->isdir = 1;
+      
+   files = g_list_prepend (files, newfle);
+   (*total)++;
    return (files);
 }
 /*****************************************************************************/
 void add_file_listbox (struct ftp_window_data *wdata, struct ftp_file_data *fle) {
    char *add_data[7] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL};
    struct pix_ext *tempext;
-   char tempstr[50], *str;
+   char *tempstr, *str;
+   int clist_num;
    size_t stlen;
 
    if (!(!show_hidden_files && *fle->file == '.' && strcmp (fle->file, "..") != 0) &&
    	gtk_pattern_match_simple (wdata->filespec, fle->file)) {
-      fle->flags |= FILE_SHOWN;
+      fle->shown = 1;
    }
    else {
-      fle->flags &= ~(FILE_SHOWN | FILE_SELECTED);
+      fle->shown = 0;
+      fle->selected = 0;
       return;
    }
    stlen = strlen (fle->file);
-   fle->clist_num = gtk_clist_append (GTK_CLIST (wdata->listbox), add_data);
-   if (fle->flags & FILE_SELECTED) {
-      gtk_clist_select_row (GTK_CLIST (wdata->listbox), fle->clist_num, 0);
+   clist_num = gtk_clist_append (GTK_CLIST (wdata->listbox), add_data);
+   if (fle->selected) {
+      gtk_clist_select_row (GTK_CLIST (wdata->listbox), clist_num, 0);
       wdata->numselected++;
    }
    if (strcmp(fle->file, "..") == 0) {
-      gtk_clist_set_pixmap (GTK_CLIST (wdata->listbox), fle->clist_num, 0, dotdot_pixmap, dotdot_mask);
+      gtk_clist_set_pixmap (GTK_CLIST (wdata->listbox), clist_num, 0, dotdot_pixmap, dotdot_mask);
    }
-   else if ((fle->flags & FILE_ISLINK) && (fle->flags & FILE_ISDIR)) {
-      gtk_clist_set_pixmap (GTK_CLIST (wdata->listbox), fle->clist_num, 0, linkdir_pixmap, linkdir_mask);
+   else if (fle->islink && fle->isdir) {
+      gtk_clist_set_pixmap (GTK_CLIST (wdata->listbox), clist_num, 0, linkdir_pixmap, linkdir_mask);
    }
-   else if (fle->flags & FILE_ISLINK) {
-      gtk_clist_set_pixmap (GTK_CLIST (wdata->listbox), fle->clist_num, 0, linkfile_pixmap, linkfile_mask);
+   else if (fle->islink) {
+      gtk_clist_set_pixmap (GTK_CLIST (wdata->listbox), clist_num, 0, linkfile_pixmap, linkfile_mask);
    }
-   else if (fle->flags & FILE_ISDIR) {
-      gtk_clist_set_pixmap (GTK_CLIST (wdata->listbox), fle->clist_num, 0, dir_pixmap, dir_mask);
+   else if (fle->isdir) {
+      gtk_clist_set_pixmap (GTK_CLIST (wdata->listbox), clist_num, 0, dir_pixmap, dir_mask);
    }
-   else if (fle->flags & FILE_ISEXE) {
-      gtk_clist_set_pixmap (GTK_CLIST (wdata->listbox), fle->clist_num, 0, exe_pixmap, exe_mask);
+   else if (fle->isexe) {
+      gtk_clist_set_pixmap (GTK_CLIST (wdata->listbox), clist_num, 0, exe_pixmap, exe_mask);
    }
    else {
       tempext = registered_exts;
@@ -188,12 +179,10 @@ void add_file_listbox (struct ftp_window_data *wdata, struct ftp_file_data *fle)
          if (stlen >= tempext->stlen &&
             strcmp (&fle->file[stlen-tempext->stlen], tempext->ext) == 0) {
 
-            if (toupper(*tempext->ascii_binary) == 'A') {
-               fle->flags |= FILE_ASCII;
-            }
+            if (toupper(*tempext->ascii_binary) == 'A') fle->ascii = 1;
             
             if (tempext->pixmap != NULL && *tempext->filename != '\0') {
-               gtk_clist_set_pixmap (GTK_CLIST (wdata->listbox), fle->clist_num, 0, tempext->pixmap, tempext->mask);
+               gtk_clist_set_pixmap (GTK_CLIST (wdata->listbox), clist_num, 0, tempext->pixmap, tempext->mask);
             }
             else tempext = NULL;
             break;
@@ -201,31 +190,31 @@ void add_file_listbox (struct ftp_window_data *wdata, struct ftp_file_data *fle)
          tempext = tempext->next;
       }
       if (tempext == NULL) {
-         gtk_clist_set_pixmap (GTK_CLIST (wdata->listbox), fle->clist_num, 0, doc_pixmap, doc_mask);
+         gtk_clist_set_pixmap (GTK_CLIST (wdata->listbox), clist_num, 0, doc_pixmap, doc_mask);
       }
    }
-   gtk_clist_set_text (GTK_CLIST (wdata->listbox), fle->clist_num, 1, fle->file);
+   gtk_clist_set_text (GTK_CLIST (wdata->listbox), clist_num, 1, fle->file);
 
    if (fle->attribs && (*fle->attribs == 'b' || *fle->attribs == 'c')) {
-      g_snprintf (tempstr, sizeof (tempstr), "%d, %d", (int) fle->size >> 16, (int) fle->size & 0xFF);
+      tempstr = g_strdup_printf ("%d, %d", (int) fle->size >> 16, (int) fle->size & 0xFF);
    }
    else {
-      insert_commas (fle->size, tempstr, sizeof (tempstr));
+      tempstr = insert_commas (fle->size);
    }
-   tempstr[sizeof (tempstr) - 1] = '\0';
-   gtk_clist_set_text (GTK_CLIST (wdata->listbox), fle->clist_num, 2, tempstr);
+   gtk_clist_set_text (GTK_CLIST (wdata->listbox), clist_num, 2, tempstr);
+   g_free (tempstr);
 
    if (fle->user) {
-      gtk_clist_set_text (GTK_CLIST (wdata->listbox), fle->clist_num, 3, fle->user);
+      gtk_clist_set_text (GTK_CLIST (wdata->listbox), clist_num, 3, fle->user);
    }
    if (fle->group) {
-      gtk_clist_set_text (GTK_CLIST (wdata->listbox), fle->clist_num, 4, fle->group);
+      gtk_clist_set_text (GTK_CLIST (wdata->listbox), clist_num, 4, fle->group);
    }
    if ((str = ctime (&fle->datetime))) {
-      gtk_clist_set_text (GTK_CLIST (wdata->listbox), fle->clist_num, 5, str);
+      gtk_clist_set_text (GTK_CLIST (wdata->listbox), clist_num, 5, str);
    }
    if (fle->attribs) {
-      gtk_clist_set_text (GTK_CLIST (wdata->listbox), fle->clist_num, 6, fle->attribs);
+      gtk_clist_set_text (GTK_CLIST (wdata->listbox), clist_num, 6, fle->attribs);
    }
 }
 /*****************************************************************************/
@@ -266,30 +255,27 @@ char *alltrim (char *str) {
    return (str);
 }
 /*****************************************************************************/
-int expand_path (char *src, char *dest, size_t size) {
-   char *prevpos, *pos, *endpos, tempchar, *newstr, *str;
+char *expand_path (char *src) {
+   char *str, *pos, *endpos, *prevpos, *newstr, *tempstr, tempchar;
    struct passwd *pw;
-   size_t len;
    
    pw = NULL;
-   if (strcmp(src, "~") == 0 || strncmp (src, "~/", 2) == 0) pw = getpwuid (geteuid ());
-   else if (strncmp(src, "~", 1) == 0) {
-      if ((pos = strchr (src, '/')) == NULL) {
-         *dest = '\0';
-         return (0);
-      }
-      tempchar = *pos;
-      *pos = '\0';
-      pw = getpwnam (src+1);
-      *pos = tempchar;
-   }
-   
-   newstr = g_malloc (size);
-   memset (newstr, 0, size);
-   len = size-1;
    str = g_malloc (strlen (src) + 1);
    strcpy (str, src);
+
+   if (*str == '~') {
+      if (*(str + 1) == '/' || (pos = strchr (str, '/')) == NULL) {
+	 pw = getpwuid (geteuid ());
+      }
+      else {
+         *pos = '\0';
+	 pw = getpwnam (str+1);
+	 *pos = '/';
+      }
+   }   
+
    endpos = str;
+   newstr = NULL;
    while ((pos = strchr (endpos, '/')) != NULL) {
       pos++;
       if ((endpos = strchr (pos, '/')) == NULL) endpos = strchr (pos, '\0');
@@ -297,39 +283,52 @@ int expand_path (char *src, char *dest, size_t size) {
       *endpos = '\0';
       if (strcmp (pos, "..") == 0) {
          *(pos-1) = '\0';
-         if ((prevpos = strrchr (newstr, '/')) != NULL) *prevpos = '\0';
+         if (newstr != NULL && (prevpos = strrchr (newstr, '/')) != NULL) *prevpos = '\0';
       }
       else if (strcmp (pos, ".") != 0) {
-         len -= strlen (pos);
-         if (len < 1) break;
-         strncat (newstr, pos-1, len);
+         if (newstr == NULL) {
+            newstr = g_malloc (strlen (pos - 1) + 1);
+            strcpy (newstr, pos - 1);
+         }
+         else {
+            tempstr = g_strconcat (newstr, pos - 1, NULL);
+            g_free (newstr);
+            newstr = tempstr;
+         }
       }
       *endpos = tempchar;
       if (*endpos == '\0') break;
-      endpos = pos+1;
-   }
-   g_free (str);
-   if (*newstr == '\0') {
-      *newstr = '/';
-      *(newstr+1) = '\0';
-   }
-   
-   if (pw != NULL) {
-      if (pos == NULL) {
-         strncpy (dest, pw->pw_dir, size);
-      }
-      else {
-         g_snprintf (dest, size, "%s%s", pw->pw_dir, newstr);
-      }
-      dest[size-1] = '\0';
-   }
-   else {
-      strncpy (dest, newstr, size);
-      dest[size-1] = '\0';
+      endpos = pos + 1;
    }
 
-   g_free (newstr);   
-   return (1);
+   if (newstr == NULL || *newstr == '\0') {
+      if (newstr != NULL) g_free (newstr);
+      newstr = g_malloc (2);
+      *newstr = '/'; 
+      *(newstr + 1) = '\0';
+   }
+
+   if (newstr != NULL) {
+      g_free (str);
+      str = newstr;
+   }
+
+   if (pw != NULL) {
+      if ((pos = strchr (str, '/')) == NULL) {
+         newstr = g_malloc (strlen (pw->pw_dir) + 1);
+         strcpy (newstr, pw->pw_dir);
+      }
+      else {
+         newstr = g_strconcat (pw->pw_dir, pos, NULL);
+      }
+      return (newstr);
+   }
+   else {
+      newstr = g_malloc (strlen (str) + 1);
+      strcpy (newstr, str);
+   }
+   g_free (str);
+   return (newstr);
 }
 /*****************************************************************************/
 void remove_double_slashes (char *string) {
@@ -345,30 +344,33 @@ void remove_double_slashes (char *string) {
    if (string[strlen (string) - 1] == '/') string[strlen (string) - 1] = '\0';
 }
 /*****************************************************************************/
-char *make_temp_filename (char *destbuf, char *ext, int size) {
-   char tempstr[MAXSTR];
+char *make_temp_filename (char *ext) {
+   char *tempstr, *destbuf;
    
-   expand_path (BASE_CONF_DIR "/temp", tempstr, sizeof(tempstr));
+   tempstr = expand_path (BASE_CONF_DIR "/temp");
    if (access (tempstr, F_OK) == -1) mkdir (tempstr, 0x1C0);
 
    srand (time (NULL));
-   g_snprintf (destbuf, size, "%s/temp%ld%s", tempstr, 1+(long) (99999999.0*rand()/(RAND_MAX+1.0)), ext == NULL ? "" : ext);
-   destbuf[size-1] = '\0';
-   while (access (destbuf, F_OK) != -1) {
-      g_snprintf (destbuf, size, "%s/temp%ld%s", tempstr, 1+(long) (99999999.0*rand()/(RAND_MAX+1.0)), ext == NULL ? "" : ext);
-      destbuf[size-1] = '\0';
-   }
+   destbuf = NULL;
+   do {
+      if (destbuf) g_free (destbuf);
+      destbuf = g_strdup_printf ("%s/temp%ld%s", tempstr, 1+(long) (99999999.0*rand()/(RAND_MAX+1.0)), ext == NULL ? "" : ext);
+   } while (access (destbuf, F_OK) != -1);
+   g_free (tempstr);
    return (destbuf);
 }
 /*****************************************************************************/
-void free_file_list (struct ftp_file_data *filelist) {
+void free_file_list (GList *filelist) {
    struct ftp_file_data *tempfle;
+   GList *templist;
    
-   while (filelist != NULL) {
-      tempfle = filelist;
-      filelist = filelist->next;
+   templist = filelist;
+   while (templist != NULL) {
+      tempfle = (struct ftp_file_data *) templist->data;
       free_fdata (tempfle);
+      templist = templist->next;
    }
+   g_list_free (filelist);
 }
 /*****************************************************************************/
 void free_fdata (struct ftp_file_data *fle) {
@@ -389,19 +391,16 @@ struct ftp_host_data *new_hdata (void) {
    struct ftp_host_data *hdata;
    
    hdata = g_malloc0 (sizeof (struct ftp_host_data));
-   hdata->files = hdata->last = NULL;
+   hdata->files = NULL;
    hdata->ftpdata = gftp_request_new ();
+   hdata->ftpdata->resolve_symlinks = resolve_symlinks;
+   gftp_protocols[0].init (hdata->ftpdata);
+   hdata->ftpdata->proxy_hosts = proxy_hosts;
+   gftp_set_proxy_config (hdata->ftpdata, proxy_config);
    gftp_set_username (hdata->ftpdata, ANON_LOGIN);
    gftp_set_password (hdata->ftpdata, emailaddr);
-   gftp_set_proxy_hostname (hdata->ftpdata, firewall_host);
-   gftp_set_proxy_username (hdata->ftpdata, firewall_username);
-   gftp_set_proxy_password (hdata->ftpdata, firewall_password);
-   gftp_set_proxy_account (hdata->ftpdata, firewall_account);
-   gftp_set_proxy_port (hdata->ftpdata, firewall_port);
-   gftp_set_data_type (hdata->ftpdata, gftp_type_binary);
+   gftp_set_data_type (hdata->ftpdata, GFTP_TYPE_BINARY);
    gftp_logging (hdata->ftpdata, 1, ftp_log, NULL);
-   hdata->protocol = strcmp (proxy_config, "http") == 0 ? http : ftp;
-   gftp_set_proxy_config (hdata->ftpdata, proxy_config);
    GFTP_SET_TRANSFER_TYPE (hdata->ftpdata, 
    	passive_transfer ? gftp_transfer_passive : gftp_transfer_active);
    return (hdata);
@@ -418,12 +417,12 @@ struct ftp_transfer_data *new_tdata (void) {
 
    tdata = g_malloc0 (sizeof (struct ftp_transfer_data));
    tdata->hdata = new_hdata ();
+   tdata->maxkbs = maxkbs;
    gftp_logging (tdata->hdata->ftpdata, 1, queue_log, NULL);
-   tdata->flags = TRANSFER_NEW; /* Signal to make this listbox entry */
-   tdata->curtrans = 0;
    tdata->dirs_to_be_made = g_malloc (sizeof (char *));
    tdata->dirs_to_be_made = NULL;
    tdata->num_dirs_to_be_made = 0;
+   tdata->stalled = 1;
    pthread_mutex_init (&tdata->mutex, NULL);
    tdata->hdata->wdata = &window2;
    return (tdata);
@@ -441,6 +440,9 @@ void copy_hdata_struct (struct ftp_host_data *hdata, struct ftp_host_data *newhd
    if ((str = GFTP_GET_PASSWORD (hdata->ftpdata))) {
       gftp_set_password (newhdata->ftpdata, str);
    }
+   if ((str = GFTP_GET_ACCOUNT (hdata->ftpdata))) {
+      gftp_set_account (newhdata->ftpdata, str);
+   }
    if ((str = GFTP_GET_DIRECTORY (hdata->ftpdata))) {
       gftp_set_directory (newhdata->ftpdata, str);
    }
@@ -454,16 +456,22 @@ void copy_hdata_struct (struct ftp_host_data *hdata, struct ftp_host_data *newhd
    if ((str = GFTP_GET_PROXY_PASSWORD (hdata->ftpdata))) {
       gftp_set_proxy_password (newhdata->ftpdata, str);
    }
+   if ((str = GFTP_GET_PROXY_ACCOUNT (hdata->ftpdata))) {
+      gftp_set_proxy_account (newhdata->ftpdata, str);
+   }
+   if ((str = GFTP_GET_PROXY_CONFIG (hdata->ftpdata))) {
+      gftp_set_proxy_config (newhdata->ftpdata, str);
+   }
    gftp_set_proxy_port (newhdata->ftpdata, GFTP_GET_PROXY_PORT (hdata->ftpdata));
    GFTP_SET_LOGGING (newhdata->ftpdata, GFTP_GET_LOGGING (hdata->ftpdata));
    gftp_set_data_type (newhdata->ftpdata, GFTP_GET_DATA_TYPE (hdata->ftpdata));
    newhdata->ftpdata->transfer_type = hdata->ftpdata->transfer_type;
    newhdata->ftpdata->logging_function = hdata->ftpdata->logging_function;
    newhdata->ftpdata->user_data = hdata->ftpdata->user_data;
-   newhdata->protocol = hdata->protocol;
 }
 /*****************************************************************************/
 void copy_fdata_struct (struct ftp_file_data *fle, struct ftp_file_data *newfle) {
+   memcpy (newfle, fle, sizeof (struct ftp_file_data ));
    if (fle->file) {
       newfle->file = g_malloc (strlen (fle->file) + 1);
       strcpy (newfle->file, fle->file);
@@ -484,14 +492,14 @@ void copy_fdata_struct (struct ftp_file_data *fle, struct ftp_file_data *newfle)
       newfle->attribs = g_malloc (strlen (fle->attribs) + 1);
       strcpy (newfle->attribs, fle->attribs);
    }
-   newfle->datetime = fle->datetime;
-   newfle->size = fle->size;
-   newfle->remote_size = fle->remote_size;
-   newfle->flags = fle->flags;
+   if (fle->remote_file) {
+      newfle->remote_file = g_malloc (strlen (fle->remote_file) + 1);
+      strcpy (newfle->remote_file, fle->remote_file);
+   }
 }
 /*****************************************************************************/
 int compare_hdata_structs (gftp_request *request1, gftp_request *request2, int compare_dirs) {
-   char *strarr[4][2];
+   char *strarr[3][2];
    int i, ret;
 
    ret = 1;
@@ -500,16 +508,14 @@ int compare_hdata_structs (gftp_request *request1, gftp_request *request2, int c
       strarr[0][1] = GFTP_GET_HOSTNAME (request2);
       strarr[1][0] = GFTP_GET_USERNAME (request1);
       strarr[1][1] = GFTP_GET_USERNAME (request2);
-      strarr[2][0] = GFTP_GET_PASSWORD (request1);
-      strarr[2][1] = GFTP_GET_PASSWORD (request2);
       if (compare_dirs) {
-         strarr[3][0] = GFTP_GET_DIRECTORY (request1);
-         strarr[3][1] = GFTP_GET_DIRECTORY (request2);
+         strarr[2][0] = GFTP_GET_DIRECTORY (request1);
+         strarr[2][1] = GFTP_GET_DIRECTORY (request2);
       }
       else {
-         strarr[3][0] = strarr[3][1] = "";
+         strarr[2][0] = strarr[2][1] = "";
       }
-      for (i=0; i<4; i++) {
+      for (i=0; i<3; i++) {
          if ((strarr[i][0] && !strarr[i][1]) || (!strarr[i][0] && strarr[i][1])) {
             ret = 0;
             break;
@@ -525,51 +531,62 @@ int compare_hdata_structs (gftp_request *request1, gftp_request *request2, int c
    return (ret);
 }
 /*****************************************************************************/
-struct ftp_file_data *get_next_selected_filename (struct ftp_file_data *filelist) {
+GList *get_next_selected_filename (GList *filelist) {
    struct ftp_file_data *tempfle;
+   GList *templist;
    
-   tempfle = filelist;
-   while (tempfle != NULL) {
-      if ((tempfle->flags & FILE_SHOWN) && (tempfle->flags & FILE_SELECTED)) {
-         return (tempfle);
+   templist = filelist;
+   while (templist != NULL) {
+      tempfle = (struct ftp_file_data *) templist->data;
+      if (tempfle->shown && tempfle->selected) {
+         return (templist);
       }
-      tempfle = tempfle->next;
+      templist = templist->next;
    }
    return (NULL);
 }
 /*****************************************************************************/
 void open_xpm (char *filename, GtkWidget *parent, GdkPixmap **pixmap, GdkBitmap **mask, int quit_on_err) {
-   char tempstr[MAXSTR], *temp1str;
+   char *exfile;
    GtkStyle *style;
 
    if (*filename == '\0') return;
    style = gtk_widget_get_style (parent);
-   temp1str = g_strconcat (BASE_CONF_DIR, "/", filename, NULL);
-   expand_path (temp1str, tempstr, sizeof (tempstr));
-   if (access (tempstr, F_OK) != 0) {
-      g_free (temp1str);
-      temp1str = g_strconcat (SHARE_DIR, "/", filename, NULL);
-      expand_path (temp1str, tempstr, sizeof (tempstr));
-      if (access (tempstr, F_OK) != 0) {
-         g_free (temp1str);
-         temp1str = g_strconcat ("/usr/share/icons/", filename, NULL);
-         expand_path (temp1str, tempstr, sizeof (tempstr));
-         if (access (tempstr, F_OK) != 0) {
-            g_free (temp1str);
-            if (!quit_on_err) return;
+
+   if ((exfile = get_xpm_path (filename, quit_on_err)) == NULL) return;
+   
+   *pixmap = gdk_pixmap_create_from_xpm (parent->window, mask, &style->bg[GTK_STATE_NORMAL], exfile);
+   if (*pixmap == NULL) {
+      if (!quit_on_err) return;
+      printf (_("gFTP Error: Error opening file %s\n"), exfile);
+      exit (-1);
+   }
+   g_free (exfile);
+}
+/*****************************************************************************/
+char *get_xpm_path (char *filename, int quit_on_err) {
+   char *tempstr, *exfile;
+   
+   tempstr = g_strconcat (BASE_CONF_DIR, "/", filename, NULL);
+   exfile = expand_path (tempstr);
+   g_free (tempstr);
+   if (access (exfile, F_OK) != 0) {
+      g_free (exfile);
+      tempstr = g_strconcat (SHARE_DIR, "/", filename, NULL);
+      exfile = expand_path (tempstr);
+      g_free (tempstr);
+      if (access (exfile, F_OK) != 0) {
+         g_free (exfile);
+         exfile = g_strconcat ("/usr/share/icons/", filename, NULL);
+         if (access (exfile, F_OK) != 0) {
+            g_free (exfile);
+            if (!quit_on_err) return (NULL);
             printf (_("gFTP Error: Cannot find file %s in %s or %s\n"), filename, SHARE_DIR, BASE_CONF_DIR);
             exit (-1);
          }
       }
    }
-   *pixmap = gdk_pixmap_create_from_xpm (parent->window, mask, &style->bg[GTK_STATE_NORMAL], tempstr);
-   if (*pixmap == NULL) {
-      g_free (temp1str);
-      if (!quit_on_err) return;
-      printf(_("gFTP Error: Error opening file %s\n"), tempstr);
-      exit (-1);
-   }
-   g_free (temp1str);
+   return (exfile);
 }
 /*****************************************************************************/
 GtkWidget *toolbar_pixmap (GtkWidget *widget, char **xpmdata) {
@@ -589,7 +606,7 @@ GtkWidget *toolbar_pixmap (GtkWidget *widget, char **xpmdata) {
 }
 /*****************************************************************************/
 int check_status(char *name, struct ftp_window_data *wdata, int only_one, 
-	int at_least_one, int use_ftp_protocol) {
+	int at_least_one, int func) {
 
    if (wdata->hdata->stopable) {
       ftp_log (gftp_logging_misc, NULL, _("%s: Please hit the stop button first to do anything else\n"), name);
@@ -601,8 +618,8 @@ int check_status(char *name, struct ftp_window_data *wdata, int only_one,
       return (0);
    }
 
-   if (use_ftp_protocol && wdata->hdata->protocol != ftp && wdata->local != 1) {
-      ftp_log (gftp_logging_misc, NULL, _("%s: You must be connected through the FTP protocol\n"), name);
+   if (!func) {
+      ftp_log (gftp_logging_misc, NULL, _("%s: This feature is not available using this protocol\n"), name);
       return (0);
    }
 
@@ -631,8 +648,50 @@ void create_item_factory (GtkItemFactory *ifactory, guint n_entries, GtkItemFact
    
    for (i=0; i<n_entries; i++) {
       memcpy (&dummy_item, entries + i, sizeof (GtkItemFactoryEntry));
+      if (dummy_item.item_type) {
+         dummy_item.item_type = _(entries[i].item_type);
+      }
       dummy_item.path = _(entries[i].path);
       gtk_item_factory_create_item (ifactory, &dummy_item, callback_data, 1);
    }
 }
 /*****************************************************************************/
+void add_history (GtkWidget *widget, GList **history, int *histlen, char *str) {
+   GList *node, *delnode;
+   char *tempstr;
+   int i;
+   
+   if (str == NULL || *str == '\0') return;
+   for (node = *history; node != NULL; node = node->next) {
+      if (strcmp ((char *) node->data, str) == 0) break;
+   }
+
+   if (node == NULL) {
+      if (*histlen >= MAX_HIST_LEN) {
+         node = *history;
+  	 for (i = 1; i < MAX_HIST_LEN; i++) node = node->next;
+         node->prev->next = NULL;
+         node->prev = NULL;
+         delnode = node;
+         while (delnode != NULL) {
+            if (delnode->data) g_free (delnode->data);
+            delnode = delnode->next;
+	 }
+         g_list_free (node);
+      }
+      tempstr = g_malloc (strlen (str) + 1);
+      strcpy (tempstr, str);
+      *history = g_list_prepend (*history, tempstr);
+      ++*histlen;
+   }
+   else if (node->prev != NULL) {
+      node->prev->next = node->next;
+      if (node->next != NULL) node->next->prev = node->prev;
+      node->prev = NULL;
+      node->next = *history;
+      if (node->next != NULL) node->next->prev = node;
+      *history = node;
+   }
+   gtk_combo_set_popdown_strings (GTK_COMBO (widget), *history);
+}
+/*****************************************************************************/
