@@ -30,10 +30,8 @@ static void dont_reconnect(GtkWidget *widget, struct dialog_data *data);
 static void dosavelog(GtkWidget *widget, GtkFileSelection *fs);
 
 void about_dialog (gpointer data) {
-   char *no_license_agreement;
-   
    GtkWidget *tempwid, *notebook, *box, *label, *view, *vscroll;
-   char tempstr[MAXSTR], *str;
+   char *tempstr, *no_license_agreement, *str, buf[MAXSTR];
    GtkWidget *dialog;
    FILE *fd;
    
@@ -46,11 +44,10 @@ void about_dialog (gpointer data) {
    gtk_box_set_homogeneous (GTK_BOX (GTK_DIALOG (dialog)->action_area), TRUE);
    gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
    gtk_signal_connect_object (GTK_OBJECT (dialog), "delete_event", GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (dialog));
-   gtk_signal_connect_object (GTK_OBJECT (dialog), "destroy", GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (dialog));
    gtk_widget_realize (dialog);
       
    notebook = gtk_notebook_new ();
-   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), notebook, TRUE, TRUE, TRUE);
+   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), notebook, TRUE, TRUE, 0);
    gtk_widget_show (notebook);
    
    box = gtk_vbox_new (TRUE, 5);
@@ -59,18 +56,18 @@ void about_dialog (gpointer data) {
    
    open_xpm ("gftp-logo.xpm", dialog, &logo_pixmap, &logo_mask, 0);
    tempwid = gtk_pixmap_new (logo_pixmap, logo_mask);
-   gtk_box_pack_start (GTK_BOX (box), tempwid, FALSE, FALSE, FALSE);
+   gtk_box_pack_start (GTK_BOX (box), tempwid, FALSE, FALSE, 0);
    gtk_widget_show (tempwid);
 
-   g_snprintf (tempstr, sizeof (tempstr), _("%s\nCopyright (C) 1998-1999 Brian Masney <masneyb@seul.org>\nOfficial Homepage: http://gftp.seul.org/\nLogo by: Aaron Worley <planet_hoth@yahoo.com>\n"), version);
-   tempstr[sizeof (tempstr) - 1] = '\0';
+   tempstr = g_strdup_printf (_("%s\nCopyright (C) 1998-1999 Brian Masney <masneyb@seul.org>\nOfficial Homepage: http://gftp.seul.org/\nLogo by: Aaron Worley <planet_hoth@yahoo.com>\n"), version);
    str = _("Translated by");
    if (strcmp (str, "Translated by") != 0) {
-      strncat (tempstr, str, sizeof (tempstr));
-      tempstr[sizeof (tempstr) - 1] = '\0';
+      tempstr = g_realloc (tempstr, strlen (tempstr) + strlen (str) + 1);
+      strcat (tempstr, str);
    }
    tempwid = gtk_label_new (tempstr);
-   gtk_box_pack_start (GTK_BOX (box), tempwid, FALSE, FALSE, FALSE);
+   g_free (tempstr);
+   gtk_box_pack_start (GTK_BOX (box), tempwid, FALSE, FALSE, 0);
    gtk_widget_show (tempwid);
    
    label = gtk_label_new (_("About"));
@@ -83,7 +80,7 @@ void about_dialog (gpointer data) {
    gtk_widget_show (box);
 
    tempwid = gtk_table_new (1, 2, FALSE);
-   gtk_box_pack_start (GTK_BOX (box), tempwid, TRUE, TRUE, TRUE);
+   gtk_box_pack_start (GTK_BOX (box), tempwid, TRUE, TRUE, 0);
    gtk_widget_show (tempwid);
 
    view = gtk_text_new (NULL, NULL);
@@ -105,16 +102,18 @@ void about_dialog (gpointer data) {
 
    tempwid = gtk_button_new_with_label (_("  Close  "));
    GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
-   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, FALSE, FALSE, FALSE);
+   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, FALSE, FALSE, 0);
    gtk_signal_connect_object (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (dialog));
    gtk_widget_grab_default (tempwid);
    gtk_widget_show (tempwid);
 
-   expand_path (SHARE_DIR "/COPYING", tempstr, sizeof (tempstr));
+   tempstr = expand_path (SHARE_DIR "/COPYING");
    if (access (tempstr, F_OK) != 0) {
-      expand_path (BASE_CONF_DIR "/COPYING", tempstr, sizeof (tempstr));
+      g_free (tempstr);
+      tempstr = expand_path (BASE_CONF_DIR "/COPYING");
       if (access (tempstr, F_OK) != 0) {
-         strncpy (tempstr, "/usr/doc/copyright/GPL", sizeof (tempstr));
+         g_free (tempstr);
+         tempstr = g_strconcat ("/usr/doc/copyright/GPL", NULL);
          if (access (tempstr, F_OK) != 0) {
             gtk_text_insert (GTK_TEXT (view), NULL, NULL, NULL, no_license_agreement, strlen (no_license_agreement));
             gtk_widget_show (dialog);
@@ -125,12 +124,14 @@ void about_dialog (gpointer data) {
    if ((fd = fopen (tempstr, "r")) == NULL) {
       gtk_text_insert (GTK_TEXT (view), NULL, NULL, NULL, no_license_agreement, strlen (no_license_agreement));
       gtk_widget_show (dialog);
+      g_free (tempstr);
       return;
    }
-   memset (tempstr, 0, sizeof (tempstr));
-   while (fread (tempstr, 1, sizeof (tempstr) - 1, fd)) {
-      gtk_text_insert (GTK_TEXT (view), NULL, NULL, NULL, tempstr, strlen (tempstr));
-      memset (tempstr, 0, sizeof (tempstr));
+   g_free (tempstr);
+   
+   memset (buf, 0, sizeof (buf));
+   while (fread (buf, 1, sizeof (buf) - 1, fd)) {
+      gtk_text_insert (GTK_TEXT (view), NULL, NULL, NULL, buf, strlen (buf));
    }
    fclose (fd);
    gtk_widget_show (dialog);
@@ -159,15 +160,15 @@ void MakeEditDialog (char *diagtxt, char *infotxt, char *deftext, int edit_shown
    gtk_box_set_homogeneous (GTK_BOX (GTK_DIALOG (d->dialog)->action_area), TRUE);
    gtk_window_set_position (GTK_WINDOW (d->dialog), GTK_WIN_POS_MOUSE);
    if (cancelfunc) {
-      gtk_signal_connect (GTK_OBJECT (d->dialog), "destroy", GTK_SIGNAL_FUNC (cancelfunc), (gpointer) cancel_d);
+      gtk_signal_connect (GTK_OBJECT (d->dialog), "delete_event", GTK_SIGNAL_FUNC (cancelfunc), (gpointer) cancel_d);
    }
-   gtk_signal_connect_object (GTK_OBJECT (d->dialog), "destroy", GTK_SIGNAL_FUNC (destroy_dialog_data), (gpointer) cancel_d);
+   gtk_signal_connect_object (GTK_OBJECT (d->dialog), "delete_event", GTK_SIGNAL_FUNC (destroy_dialog_data), (gpointer) cancel_d);
    tempwid = gtk_label_new (infotxt);
-   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (d->dialog)->vbox), tempwid, TRUE, TRUE, FALSE);
+   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (d->dialog)->vbox), tempwid, TRUE, TRUE, 0);
    gtk_widget_show (tempwid);
    
    d->edit = cancel_d->edit = gtk_entry_new ();
-   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (d->dialog)->vbox), d->edit, TRUE, TRUE, FALSE);
+   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (d->dialog)->vbox), d->edit, TRUE, TRUE, 0);
    if (okfunc) {
       gtk_signal_connect (GTK_OBJECT (d->edit), "activate", GTK_SIGNAL_FUNC (okfunc), (gpointer) d);
    }
@@ -184,7 +185,7 @@ void MakeEditDialog (char *diagtxt, char *infotxt, char *deftext, int edit_shown
 
    tempwid = gtk_button_new_with_label (oktxt);
    GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
-   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (d->dialog)->action_area), tempwid, TRUE, TRUE, FALSE);
+   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (d->dialog)->action_area), tempwid, TRUE, TRUE, 0);
    if (okfunc) {
       gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (okfunc), (gpointer) d);
    }
@@ -196,7 +197,7 @@ void MakeEditDialog (char *diagtxt, char *infotxt, char *deftext, int edit_shown
             
    tempwid = gtk_button_new_with_label (canceltxt);
    GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
-   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (d->dialog)->action_area), tempwid, TRUE, TRUE, FALSE);
+   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (d->dialog)->action_area), tempwid, TRUE, TRUE, 0);
    if (cancelfunc) {
       gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC(cancelfunc), (gpointer) cancel_d);
    }
@@ -222,7 +223,6 @@ void MakeYesNoDialog (char *diagtxt, char *infotxt, int erase, int num, ...) {
    void *ptr;
    int i;
    
-/*FIXME*/
    dialog = gtk_dialog_new ();
    gtk_grab_add (dialog);
    gtk_window_set_title (GTK_WINDOW (dialog), diagtxt);
@@ -234,7 +234,7 @@ void MakeYesNoDialog (char *diagtxt, char *infotxt, int erase, int num, ...) {
    gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
    
    text = gtk_label_new (infotxt);
-   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), text, TRUE, TRUE, FALSE);
+   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), text, TRUE, TRUE, 0);
    gtk_widget_show (text);
 
    list = NULL;
@@ -252,7 +252,7 @@ void MakeYesNoDialog (char *diagtxt, char *infotxt, int erase, int num, ...) {
       d->all_buttons = list;
       
       tempwid = gtk_button_new_with_label (tempstr);
-      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, TRUE);
+      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid, TRUE, TRUE, 0);
       if (myfunc) {
          gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (myfunc), (gpointer) d);
       }
@@ -278,7 +278,7 @@ void destroy_dialog_data (struct dialog_data *data) {
 }
 /*****************************************************************************/
 void site_dialog (struct ftp_window_data *wdata) {
-   if (!check_status (_("Site"), wdata, 0, 0, 1)) return;
+   if (!check_status (_("Site"), wdata, 0, 0, wdata->hdata->ftpdata->site != NULL)) return;
          
    MakeEditDialog (_("Site"), _("Enter site-specific command"), NULL, 1, 1,
    	_("OK"), dosite, wdata, 
@@ -300,7 +300,7 @@ static void dosite (GtkWidget *widget, struct dialog_data *data) {
 }
 /*****************************************************************************/
 void change_filespec (struct ftp_window_data *wdata) {
-   if (!check_status (_("Change Filespec"), wdata, 0, 0, 0)) return;
+   if (!check_status (_("Change Filespec"), wdata, 0, 0, 1)) return;
 
    if (wdata->local == -1) {
       ftp_log (gftp_logging_misc, NULL, _("Change Filespec: Not connected to a remote site\n"));
@@ -314,6 +314,7 @@ void change_filespec (struct ftp_window_data *wdata) {
 static void dochange_filespec (GtkWidget *widget, struct dialog_data *data) {
    struct ftp_file_data *tempfle;
    struct ftp_window_data *wdata;
+   GList *templist;
    char *edttext;
 
    wdata = (struct ftp_window_data *) data->data;
@@ -329,25 +330,28 @@ static void dochange_filespec (GtkWidget *widget, struct dialog_data *data) {
 
    gtk_clist_freeze (GTK_CLIST (wdata->listbox));
    gtk_clist_clear (GTK_CLIST (wdata->listbox));
-   tempfle = wdata->hdata->files;
-   while (tempfle != NULL) {
+   templist = wdata->hdata->files;
+   while (templist != NULL) {
+      tempfle = (struct ftp_file_data *) templist->data;
       add_file_listbox (wdata, tempfle);
-      tempfle = tempfle->next;
+      templist = templist->next;
    }
    gtk_clist_thaw (GTK_CLIST (wdata->listbox));
    update_ftp_info (wdata);
 }
 /*****************************************************************************/
-void openurl_dialog (gpointer data) {
+void tb_openurl_dialog (gpointer data) {
    char *edttxt;
    
-   edttxt = gtk_entry_get_text (GTK_ENTRY (hostedit));
+   edttxt = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (hostedit)->entry));
    if (window2.local != -1) disconnect (&window2);
    else if (edttxt != NULL && *edttxt != '\0') toolbar_hostedit (NULL, NULL);
-   else {
-      MakeEditDialog (_("Connect via URL"), _("Enter ftp url to connect to"), NULL, 1, 0,
-      		_("Connect"), do_openurl, &window2, _("  Cancel  "), NULL, NULL);
-   }
+   else openurl_dialog (data);
+}
+/*****************************************************************************/
+void openurl_dialog (gpointer data) {
+   MakeEditDialog (_("Connect via URL"), _("Enter ftp url to connect to"), NULL, 1, 1,
+      	_("Connect"), do_openurl, &window2, _("  Cancel  "), NULL, NULL);
 }
 /*****************************************************************************/
 static void do_openurl (GtkWidget *widget, struct dialog_data *data) {
@@ -357,8 +361,13 @@ static void do_openurl (GtkWidget *widget, struct dialog_data *data) {
    wdata = (struct ftp_window_data *) data->data;
    tempstr = gtk_entry_get_text (GTK_ENTRY (data->edit));
    if (gftp_parse_url (wdata->hdata->ftpdata, tempstr) == 0) {
+      if (strcmp (GFTP_GET_USERNAME (wdata->hdata->ftpdata), ANON_LOGIN) == 0) {
+         gftp_set_password (wdata->hdata->ftpdata, emailaddr);
+      }
       gtk_widget_destroy (data->dialog);
-      if (GFTP_GET_PASSWORD (wdata->hdata->ftpdata) == NULL) {
+      if (strcmp (GFTP_GET_USERNAME (wdata->hdata->ftpdata), ANON_LOGIN) == 0 && 
+      	GFTP_GET_PASSWORD (wdata->hdata->ftpdata) == NULL) {
+      	
          gftp_set_password (wdata->hdata->ftpdata, emailaddr);
       }
       if (wdata->local != -1) disconnect (wdata);
@@ -370,19 +379,6 @@ static void do_openurl (GtkWidget *widget, struct dialog_data *data) {
    }
 }
 /*****************************************************************************/
-void openurl_get_drag_data (GtkWidget *widget, GdkDragContext *context, gint x, 
-	gint y, GtkSelectionData *selection_data, guint info, guint32 clk_time, 
-	struct ftp_window_data *wdata) {
-
-   if ((selection_data->length >= 0) && (selection_data->format == 8)) {
-      if (gftp_parse_url (window2.hdata->ftpdata, selection_data->data) == 0) {
-         if (window2.local != -1) disconnect (&window2);
-         ftp_connect (window2.hdata, 1);
-      }
-      else ftp_log (gftp_logging_misc, NULL, _("Could not parse URL %s\n"), selection_data->data);
-   }
-}
-/*****************************************************************************/
 void reconnect_dialog (struct ftp_host_data *hdata) {
    char *tempstr;
 
@@ -410,13 +406,14 @@ static void dont_reconnect (GtkWidget *widget, struct dialog_data *data) {
 }
 /*****************************************************************************/
 void viewlog (gpointer data) {
-   char tempstr[MAXSTR], *txt;
+   char *tempstr, *txt;
    guint textlen;
    FILE *fd;
 
-   make_temp_filename (tempstr, NULL, sizeof (tempstr));
+   tempstr = make_temp_filename (NULL);
    if ((fd = fopen (tempstr, "w")) == NULL) {
       ftp_log (gftp_logging_misc, NULL, _("Error: Cannot open %s: %s\n"), tempstr, g_strerror (errno));
+      g_free (tempstr);
       return;
    }
    textlen = gtk_text_get_length (GTK_TEXT (logwdw));
@@ -426,6 +423,7 @@ void viewlog (gpointer data) {
    }
    fclose (fd);
    view_file (tempstr, 1, 1, 0, NULL);
+   g_free (tempstr);
 }
 /*****************************************************************************/
 void savelog (gpointer data) {
